# encoding: utf-8

# CleanUpTickets.rb
# Written by: T Lillington
# Date: 12 May 2016     Copyright 2016
# Version 2.4
#
# This program will process the Spiceworks tickets and comments tables cleaning up the tickets as it goes
#
# 1)	Remove extra blank lines from tickets and comments
# 2)	Removed unwanted signature / text blocks from tickets and comments
# 3)	Remove Attachments based on MD5 
#
require 'find'
require 'digest/md5'
require 'logger'

$LOAD_PATH << '..\..\..\..\pkg\gems\sqlite3-1.3.8\lib'

require 'sqlite3'

class CLEAN_UP_TICKETS

#########################
# Define global variables
#
  def initialize
    
    @@spice_db_name = '..\..\..\..\db\spiceworks_prod.db'
    @@ticket_folder = '..\\..\\Ticket'
    
    @@log_level = "DEBUG"
    
    @@last_blank_check_id_tickets  = 0
    @@last_blank_check_id_comments = 0
    
    @@last_text_check_id_tickets   = 0
    @@last_text_check_id_comments  = 0
    
    @@last_attachment_check_id_comments = 0
    
    @@last_blank_check_id_tickets_orig  = 0
    @@last_blank_check_id_comments_orig = 0
    
    @@last_text_check_id_tickets_orig   = 0
    @@last_text_check_id_comments_orig  = 0
    
    @@last_attachment_check_id_comments_orig = 0
  end
  
  ##############################  
  # Write out configuration file
  #
  def write_config_file()
   
    @@log.debug('Starting: Writing config file')

    if ! ( @@last_blank_check_id_tickets_orig       == @@last_blank_check_id_tickets and 
           @@last_blank_check_id_comments_orig      == @@last_blank_check_id_comments and    
           @@last_text_check_id_tickets_orig        == @@last_text_check_id_tickets and 
           @@last_text_check_id_comments_orig       == @@last_text_check_id_comments and    
           @@last_attachment_check_id_comments_orig == @@last_attachment_check_id_comments )
    
      file_contents = "# CleanUpTickets configuration settings\n\n"
    
      file_contents = file_contents + "# Don't add extra text it will not be there after this script has run\n\n"
   
      file_contents = file_contents + "# You can set the level of logging using 'DEBUG' for all messages, 'INFO' for basic day to day messages\n"
      file_contents = file_contents + "@@log_level='#{@@log_level}'\n\n"
   
      file_contents = file_contents + "# These specify the last ticket number / comment id the app checked when removing unwanted blank lines\n"
      file_contents = file_contents + "# You should not set this back or it could have an impact on subsequent text matching\n"
      file_contents = file_contents + "@@last_blank_check_id_tickets=#{@@last_blank_check_id_tickets}\n"
      file_contents = file_contents + "@@last_blank_check_id_comments=#{@@last_blank_check_id_comments}\n\n"
   
      file_contents = file_contents + "# These specify the last ticket number / comment id the app checked when looking for unwanted text\n"
      file_contents = file_contents + "# if you add additional text to remove then you can set these back to 0 and it will reprocess all\n"
      file_contents = file_contents + "# tickets/comments again\n"
      file_contents = file_contents + "@@last_text_check_id_tickets=#{@@last_text_check_id_tickets}\n"
      file_contents = file_contents + "@@last_text_check_id_comments=#{@@last_text_check_id_comments}\n\n"
   
      file_contents = file_contents + "# This specifies the last comment id the app checked when looking for unwanted attachments\n"
      file_contents = file_contents + "# if you add additional attachments to remove then you can set these back to 0\n"
      file_contents = file_contents + "# and it will reprocess all comments again\n"
      file_contents = file_contents + "@@last_attachment_check_id_comments=#{@@last_attachment_check_id_comments}\n"

       
      # Create a new file
      File.open('CleanUpTickets.ini', 'w') do |filecontents|
        filecontents.puts "#{file_contents}"
      end
    end
    
    @@log.debug('Ending  : Writing config file')
    
  end
  
  #########################   
  # Read configuration data
  #
  def read_config_file()
    
    #@@log.debug('Starting: Read config file')
   
    if not File.exists?('CleanUpTickets.ini')
      @@log.info('Config file not found, creating it')
      @@last_blank_check_id_tickets_orig = 1    # make write function think the settings have changed
      write_config_file()
    end
    
    # execute ini file to create variables from parameter
    load 'CleanUpTickets.ini'
 
    # Remember original config parameters
    @@last_blank_check_id_tickets_orig  = @@last_blank_check_id_tickets
    @@last_blank_check_id_comments_orig = @@last_blank_check_id_comments
    
    @@last_text_check_id_tickets_orig   = @@last_text_check_id_tickets
    @@last_text_check_id_comments_orig  = @@last_text_check_id_comments
    
    @@last_attachment_check_id_comments_orig = @@last_attachment_check_id_comments
    
    #@@log.debug('Ending  : Read config file')
    
  end
  
  #########################################################   
  # Load the MD5 values used to identify the unwanted files 
  #
  def load_file_md5s_to_delete()
    
    @@log.debug('Starting: Load file MD5s to delete')
 
    @@files_to_remove_size=[]
    @@files_to_remove_MD5=[]
    
    if not File.exists?("FilesToDelete.txt")
      @@log.info('The "FilesToDelete.txt" does not exist, no attachments will be removed')
      @@log.debug('Ending  : Load file MD5s to remove')
      return
    end
    
    file_ref = File.new("FilesToDelete.txt", "r")
    
    while (line = file_ref.gets)
      # record format is size, md5 of whole file, file name
      @@files_to_remove_size << line.chomp.split("\t")[0].to_i 
      @@files_to_remove_MD5  << line.chomp.split("\t")[1]
    end
    
    file_ref.close

    @@log.debug('Ending  : Load file MD5s to delete')
   
  end
  
  ###################################################
  # scan comments table removing unwanted attachments
  #
  def remove_unwanted_attachments()
   
    @@log.debug('Starting: Remove unwanted attachments')
      
    if @@files_to_remove_size.empty?
      @@log.info('No MD5s loaded so skipping attachment removal')
      @@log.debug('Ending  : Remove unwanted attachments')
      return
    end
   
    # Open database connection
    spice_db = SQLite3::Database.open( "#{@@spice_db_name}" )
    spice_db.busy_timeout = 1000

    total_bytes_deleted  = 0
    total_files_deleted  = 0
    total_dirs_deleted   = 0
	total_thumbs_bytes_deleted = 0
	total_thumbs_files_deleted = 0
    
    # Need to filter on body as some comment entries have attachments as well, these are added by help desk staff
    # records with just "attachment:" in the body can be deleted without breaking the tickets comment trail
    while 1 == 1
      spice_db_row = spice_db.execute( "select id , ticket_id, attachment_name from comments \
                                                where id > #{@@last_attachment_check_id_comments} \
                                                and Body = 'Attachment:' and attachment_name != '' ORDER BY ID LIMIT 1")
        
      break if spice_db_row.empty?
        
      db_id              = spice_db_row[0][0]
      db_ticket_id       = spice_db_row[0][1]
      db_attachment_name = spice_db_row[0][2]
        
      @@log.debug("Processing Ticket id #{db_ticket_id.to_s} comment id #{db_id.to_s} attachment = #{db_attachment_name.to_s}")
     
      file_to_check = "#{@@ticket_folder}\\#{db_ticket_id.to_s}\\#{db_id.to_s}-#{db_attachment_name.to_s}"
	  file_thumb_name = "#{File.basename(db_attachment_name.to_s,File.extname(db_attachment_name.to_s))}.png"
      file_thumb_to_delete = "#{@@ticket_folder}\\#{db_ticket_id.to_s}\\#{db_id.to_s}-thumb-#{file_thumb_name}"
	  
      # Check file actually exists on disk
      if File.exists?(file_to_check)
        size_in_bytes = File.size(file_to_check).to_i
      
        # Search array to see if we have a file size that matches
        found_size_match = @@files_to_remove_size.index(size_in_bytes)   
      
        if found_size_match
          # We have found a file that matches the size of one of our to remove files
          # now confirm this is the one we are looking for by doing an md5 on it
         
          # Need to read in binary mode, default mode is text
          file_md5 = Digest::MD5.hexdigest(File.open(file_to_check,"rb"){|io| io.read})
         
          if file_md5 == @@files_to_remove_MD5[found_size_match]
            delete_file = 'Y'
          else
              delete_file = 'N'
              for loop_count in found_size_match + 1...@@files_to_remove_size.length
                if size_in_bytes == @@files_to_remove_size[loop_count] and file_md5 == @@files_to_remove_MD5[loop_count]
                  delete_file = 'Y'
                  break
                end
              end
          end
           
          if delete_file == 'Y'  
            @@log.debug("Deleting Ticket id #{db_ticket_id.to_s} comments record #{db_id.to_s} and associated file: #{file_to_check}")
			@@log.debug("Deleting associated thumbnail file: #{file_thumb_to_delete}")
            
            if @@log.level != 0   # only update if not debug
              begin
                  spice_db.execute( "delete from comments where id = #{db_id}" )
              rescue SQLite3::Exception => error_msg
                  @@log.info("Error while deleting attachment, comment record id = #{db_id} error message is :- #{error_msg}")
                  #Close Spiceworks database
                  spice_db.close()
                  return
              end

              begin
                  File.delete(file_to_check)
                  total_bytes_deleted = total_bytes_deleted + size_in_bytes
                  total_files_deleted = total_files_deleted + 1
              rescue StandardError => error_msg
                  @@log.info("Error deleting file: #{file_to_check} error message is :- #{error_msg}")
              end                                               

              begin
				  size_in_bytes_thumb = File.size(file_thumb_to_delete).to_i
                  File.delete(file_thumb_to_delete)
				  total_thumbs_bytes_deleted = total_thumbs_bytes_deleted + size_in_bytes_thumb
                  total_thumbs_files_deleted = total_thumbs_files_deleted + 1
              rescue StandardError => error_msg
                  @@log.info("Error deleting file: #{file_thumb_to_delete} error message is :- #{error_msg}")
              end 
			  
              # Don't want thumbs.db
              if File.exists?("#{@@ticket_folder}\\#{db_ticket_id.to_s}\\Thumbs.db")
                begin
                  File.delete("#{@@ticket_folder}\\#{db_ticket_id.to_s}\\Thumbs.db")
                rescue StandardError => error_msg
                  @@log.info("Error deleting thumbs.db file: #{@@ticket_folder}\\#{db_ticket_id.to_s}\\Thumbs.db error message is :- #{error_msg}")
                end
              end
            
              # Check to see if folder is empty now we have deleted a file
              dir_object = Dir.open("#{@@ticket_folder}\\#{db_ticket_id.to_s}")
           
              # Is folder now empty? position 2 in array is third dir entry after . and .. if empty then dir also empty
              if dir_object.seek(2).read.nil?
                @@log.debug("Deleting empty folder: #{@@ticket_folder}\\#{db_ticket_id.to_s}")
                begin
                  Dir.delete("#{@@ticket_folder}\\#{db_ticket_id.to_s}")
                  total_dirs_deleted = total_dirs_deleted + 1
                rescue StandardError => error_msg
                  @@log.info("Error deleting empty folder: #{@@ticket_folder}\\#{db_ticket_id.to_s} error message is :- #{error_msg}")
                end
              end
            
              dir_object.close
            end
          end
        end
      else
        @@log.info("Warning: The file referenced in the database not found on disk, file = #{file_to_check}")
      end
      
      @@last_attachment_check_id_comments = db_id
    end

    #Close Spiceworks database
    spice_db.close()                       
    
    if total_files_deleted != 0 or total_bytes_deleted != 0 or total_dirs_deleted != 0
      @@log.info("Deleted #{total_files_deleted} files totalling #{total_bytes_deleted} bytes and deleted #{total_dirs_deleted} empty directories")
	  @@log.info("Deleted #{total_thumbs_files_deleted} thumbnail files totalling #{total_thumbs_bytes_deleted} bytes")
	end
  
    @@log.debug('Ending  : Remove unwanted attachments')
   
  end
  
  #########################################################   
  # scan tickets and comments removing unwanted blank lines
  #
  def scan_db_remove_blank_lines()
   
    @@log.debug('Starting: Remove blank lines')
       
    # Open database connection
    spice_db = SQLite3::Database.open( "#{@@spice_db_name}" )
    spice_db.busy_timeout=1000

    db_table    = []
    db_field    = []
    db_start_id = []
    
    db_table[0]    = 'tickets'
    db_field[0]    = 'description'     
    db_start_id[0] = @@last_blank_check_id_tickets
  
    db_table[1]    = 'comments'
    db_field[1]    = 'body' 
    db_start_id[1] = @@last_blank_check_id_comments
  
    for loop_count in 0...db_table.length

      while 1 == 1
        begin
           spice_db_row = spice_db.execute "select id , #{db_field[loop_count]} from #{db_table[loop_count]} where id > #{db_start_id[loop_count]} and #{db_field[loop_count]} != '' ORDER BY ID LIMIT 1"
         rescue StandardError => error_msg
           @@log.info("Error reading from '#{db_table[loop_count]}' table, record id > #{db_start_id[loop_count]} , error message is :- #{error_msg}")
           @@log.info("If the above error is 'cant convert nil into String' and only occurs intermittently then you can ignore it.")
           db_start_id[loop_count] = db_start_id[loop_count] + 1
           retry
        end

        break if spice_db_row.empty?

        db_id   = spice_db_row[0][0]
        db_text = spice_db_row[0][1]
         
        @@log.debug("Processing '#{db_table[loop_count]}' table, record #{db_id} ( blank line removal )")

        db_text_original = db_text

        # change various "return" and "newline" combinations to single return+newline
        # did not use \r \n as I found occasional inconsistencies
        db_text = db_text.gsub(/\x0A\x0A/,'##rn##')     # double line feeds
        db_text = db_text.gsub(/\x0D\x0D/,'##rn##')     # double carriage return
        db_text = db_text.gsub(/\x0A\x0D\x0A/,'##rn##') # LF+CR+LF
        db_text = db_text.gsub(/\x0A\x0D/,'##rn##')     # LF+CR
        db_text = db_text.gsub(/\x0D\x0A/,'##rn##')     # CR+LF
        db_text = db_text.gsub(/\x0A/,'##rn##')         # LF
        db_text = db_text.gsub(/\x0D/,'##rn##')         # CR

        # where there are only spaces or tabs between return+newline entries then remove, do twice as not all combinations caught first time
        db_text = db_text.gsub(/##rn##[ \t]+##rn##/,'##rn####rn##')
        db_text = db_text.gsub(/##rn##[ \t]+##rn##/,'##rn####rn##')
        
        # where there are 3 or more return+newline's together then change to 2 lots of return+newline
        db_text = db_text.gsub(/(##rn##){3,}/,'##rn####rn##')

        # replace substitutions with crlf
        db_text = db_text.gsub('##rn##',"\r\n")

        db_text = db_text.strip
 
        #@@log.debug("Original Comment   ####################################################")
        #@@log.debug("#{db_text_original}")
        @@log.debug("Amended Comment    ----------------------------------------------------")
        @@log.debug("#{db_text}")
  
        # only update database record if its changed
        if db_text != db_text_original
          # need to translate single quotes to 2x single so db update works
          db_text = db_text.gsub("'","''")
 
          @@log.debug("Updating '#{db_table[loop_count]}' table, record #{db_id} ( blank line removal )")
 
          #if @@log.level != 0   # only update if not debug
            begin
                spice_db.execute("update #{db_table[loop_count]} set #{db_field[loop_count]} = '" + db_text + "' where id = #{db_id}")
            rescue SQLite3::Exception => error_msg
                @@log.info("Error in blank line removal while updating #{db_table[loop_count]} table record id #{db_id} error message is :- #{error_msg}")
                #Close spiceworks database
                spice_db.close()
                @@last_blank_check_id_tickets  = db_start_id[0]
                @@last_blank_check_id_comments = db_start_id[1]
                return
            end

          #end
        end

        db_start_id[loop_count] = db_id
      end
    end
   
    #Close spiceworks database
    spice_db.close()                       
   
    @@last_blank_check_id_tickets  = db_start_id[0]
    @@last_blank_check_id_comments = db_start_id[1]
   
    @@log.debug('Ending  : Remove blank lines')
   
  end
  
  ##########################################################   
  # Load the text files containing the text blocks to remove
  #
  def load_text_to_delete()
     
    @@log.debug('Starting: Load text to delete')
   
    if !File.directory?("UnwantedTextBlocks")
      Dir.mkdir("UnwantedTextBlocks")
    end
    
    @@text_to_remove = []
    @@text_to_remove_min_length = 99999999999
    
    Find.find("UnwantedTextBlocks") do |dir_entry|
      
      next if File.size(dir_entry) < 5    # Don't load empty or very small files
      
      # Only process files not folders
      if File.file?(dir_entry)
        
        @@text_to_remove << IO.binread(dir_entry)
        
        @@text_to_remove[-1].encode!("UTF-8","Windows-1252").strip    # convert incoming text file that's in Windows-1252 to UTF-8 used within SW
        if File.basename(dir_entry)[0,6] != "REGEX-"
          @@text_to_remove[-1] = @@text_to_remove[-1].gsub(/\s*/,'')   # Remove all cr/lf/tab/space
          @@text_to_remove_min_length = [@@text_to_remove_min_length,@@text_to_remove[-1].length].min
          split2one = @@text_to_remove[-1].scan(/./)    # Break string into array of 1 characters
          for temp_cnt in 0...split2one.length
            split2one[temp_cnt] = Regexp.escape(split2one[temp_cnt])
          end
          @@text_to_remove[-1] = split2one.join("\\s*")    # Add regex to match a string split anywhere with space/tab/cr/lf
          @@text_to_remove[-1] = Regexp.new(@@text_to_remove[-1])
        else    # we have a file containing a REGEX so don't process
          @@text_to_remove_min_length = [@@text_to_remove_min_length,@@text_to_remove[-1].length / 3].min
          @@text_to_remove[-1] = Regexp.new(@@text_to_remove[-1])
        end
      end
    end
    
    # sort text blocks in reverse size order so bigger blocks are removed first if not then small block removal may stop a larger block being removed
    @@text_to_remove.sort_by!{ |s| -s.source.length }

    for temp_cnt in 0...@@text_to_remove.length
      @@log.debug("Text block #{temp_cnt} loaded for removal")
      @@log.debug("#{@@text_to_remove[temp_cnt].source}")
    end
    
    @@log.debug('Ending  : Load text to delete')
   
  end
  
  ###############################################################
  # scan tickets and comments removing unwanted signatures / text
  #
  def scan_db_remove_text()
      
    @@log.debug('Starting: Remove unwanted text')
    
    if @@text_to_remove.empty?
      @@log.debug('No text for removal loaded so skipping text removal')
      @@log.debug('Ending  : Remove unwanted text')
      return
    end
   
    # Open database connection
    spice_db = SQLite3::Database.open( "#{@@spice_db_name}" )
    spice_db.busy_timeout=1000
    
    db_table    = []
    db_field    = []
    db_start_id = []
    db_end_id   = []
    
    db_table[0]    = 'tickets'
    db_field[0]    = 'description'     
    db_start_id[0] = @@last_text_check_id_tickets 
    db_end_id[0]   = @@last_blank_check_id_tickets

    db_table[1]    = 'comments'
    db_field[1]    = 'body' 
    db_start_id[1] = @@last_text_check_id_comments
    db_end_id[1]   = @@last_blank_check_id_comments

    used_text_block = []
    
    for loop_count in 0...db_table.length
      last_db_id = db_start_id[loop_count]
      
      while 1 == 1
        begin
        # only check fields that are longer than the minimum text we are looking for
        # Must not try and remove text blocks if blank line removal has not been done
        spice_db_row = spice_db.execute( "select id , #{db_field[loop_count]} from #{db_table[loop_count]}
                               where id > #{db_start_id[loop_count]} and
                                     id <= #{db_end_id[loop_count]}  and
                                     length(#{db_field[loop_count]}) > #{@@text_to_remove_min_length} and
                                     #{db_field[loop_count]} != '' ORDER BY ID LIMIT 1")   # probably don't need this as previous line should stop blank entries from being selected
         rescue StandardError => error_msg
          @@log.info("Error reading from '#{db_table[loop_count]}' table, record id > #{db_start_id[loop_count]} , error message is :- #{error_msg}")
          @@log.info("If the above error is 'cant convert nil into String' and only occurs intermittently then you can ignore it.")
          db_start_id[loop_count] = db_start_id[loop_count] + 1
          retry
        end

        break if spice_db_row.empty?

        db_id   = spice_db_row[0][0]
        db_text = spice_db_row[0][1]

        db_text_original = db_text

        @@log.debug("Processing '#{db_table[loop_count]}' table, record #{db_id} ( text block removal )")
        #@@log.debug("Original Comment   ####################################################")
        #@@log.debug("#{db_text_original}")

        #if @@log.level == 0   # if in debug mode then show workings
        #  for text_to_remove_count in 0...@@text_to_remove.length
        #    db_text_p = db_text
        #    db_text = db_text.gsub(@@text_to_remove[text_to_remove_count],"[[[[[[[[[[[[[[[[[[ Deleted text block #{text_to_remove_count} ]]]]]]]]]]]]]]]]]]\r\n")
        #    if db_text != db_text_p
        #      used_text_block[text_to_remove_count] = "Y"
        #    end
        #  end
        # else    # not in debug so delete text
          for text_to_remove_count in 0...@@text_to_remove.length
            db_text = db_text.gsub(@@text_to_remove[text_to_remove_count],"")
          end
        #end
 
        # where there are 3 or more return+newline's together then change to 2 lots of return+newline.  This can occur after text removal.
        db_text = db_text.gsub(/(\x0D\x0A){3,}/,"\r\n\r\n")
        db_text = db_text.strip
          
        @@log.debug("Amended Comment    ----------------------------------------------------")
        @@log.debug("#{db_text}")

        if db_text != db_text_original
          # need to translate single quotes to 2x single so db update works
          db_text = db_text.gsub("'","''")
            
          @@log.debug("Updating '#{db_table[loop_count]}' table, record #{db_id} ( text block removal )")
 
          if @@log.level != 0   # only update if not debug
            begin
                spice_db.execute("update #{db_table[loop_count]} set #{db_field[loop_count]} = '" + db_text + "' where id = #{db_id}")
             rescue SQLite3::Exception => error_msg
                @@log.info("Error in text block removal while updating #{db_table[loop_count]} table record id #{db_id} error message is :- #{error_msg}")
                #Close spiceworks database
                spice_db.close()
                @@last_text_check_id_tickets  = db_start_id[0]
                @@last_text_check_id_comments = db_start_id[1]
                return
            end
          end
        end
       
        db_start_id[loop_count] = db_id
      end
    end
    
    if @@log.level != 0   # only loop if in debug
      for xx in 0...used_text_block.length
        @@log.debug("#{xx}   #{used_text_block[xx]}")
      end
    end
    
    #Close spiceworks database
    spice_db.close()                       
   
    @@last_text_check_id_tickets  = db_start_id[0]
    @@last_text_check_id_comments = db_start_id[1]
   
    @@log.debug('Ending  : Remove unwanted text')
    
  end
  
  ############################# 
  # Create log file of activity
  #
  def setup_logging
    
    @@log = Logger.new( '..\..\..\..\log\CleanUpTickets.log', 10, 10000000 )
    @@log.level = Logger::DEBUG
    
  end

  ######################################################################## 
  # Set logging level to that specified in the INI file or default to INFO
  # 0=debug   1=info   2=warn   3=error   4=fatal
  def set_logging_level
    
    case @@log_level.upcase
      when 'INFO'
        @@log.level = Logger::INFO
    end
    
    $log_level = @@log.level
  
  end

  ##################
  # Turn off logging
  #
  def stop_logging
    
    @@log.close()
    
  end

  ###########################   
  # Add log entry to log file
  #
  def log_message(msg_text)
    
    @@log.info("#{msg_text}")
    
  end
end

##############
# Main section
#

clean_up = CLEAN_UP_TICKETS.new

clean_up.setup_logging

#clean_up.log_message('Running: Ticket Cleanup')

# Read where to continue processing from 
clean_up.read_config_file()

# Reset logging level to that specified in the config file
clean_up.set_logging_level

# Remove unwanted attachments/files
clean_up.load_file_md5s_to_delete()
clean_up.remove_unwanted_attachments()

# Scan all tickets and comments removing unwanted blank lines
clean_up.scan_db_remove_blank_lines()

# Scan all tickets and comments removing unwanted text
clean_up.load_text_to_delete()
clean_up.scan_db_remove_text()

# save details of what we have just processed if not in debug mode
if $log_level != 0
  clean_up.write_config_file()
end

clean_up.log_message('Ending  : Ticket Cleanup')

clean_up.stop_logging
